@@ -84,6 +84,12 @@ template<> struct hash_ops<int64_t> : hash_int_ops
 		return mkhash((unsigned int)(a), (unsigned int)(a >> 32));
 	}
 };
+template<> struct hash_ops<uint32_t> : hash_int_ops
+{
+	static inline unsigned int hash(uint32_t a) {
+		return a;
+	}
+};
 
 template<> struct hash_ops<std::string> {
 	static inline bool cmp(const std::string &a, const std::string &b) {
@@ -189,7 +195,7 @@ inline int hashtable_size(int min_size)
 		if (p >= min_size) return p;
 
 	if (sizeof(int) == 4)
-		throw std::length_error("hash table exceeded maximum size. use a ILP64 abi for larger tables.");
+		throw std::length_error("hash table exceeded maximum size.\nDesign is likely too large for yosys to handle, if possible try not to flatten the design.");
 
 	for (auto p : zero_and_some_primes)
 		if (100129 * p > min_size) return 100129 * p;
@@ -982,7 +988,7 @@ public:
 		return !operator==(other);
 	}
 
-	bool hash() const {
+	unsigned int hash() const {
 		unsigned int hashval = mkhash_init;
 		for (auto &it : entries)
 			hashval ^= ops.hash(it.udata);
